local _, addon = ...;

FactionBroker = LibStub("AceAddon-3.0"):NewAddon(addon, "FactionBroker", "AceEvent-3.0");

local L = LibStub("AceLocale-3.0"):GetLocale("FactionBroker");
local LSM = LibStub("LibSharedMedia-3.0");

-- Upvalue important globals
local GetFactionInfo, GetNumFactions, GetWatchedFactionInfo, SetWatchedFactionIndex = GetFactionInfo, GetNumFactions, GetWatchedFactionInfo, SetWatchedFactionIndex;
local GENDER_NEUTRAL, GENDER_MALE, GENDER_FEMALE = 1, 2, 3; -- UnitSex return values
local FactionStandingText = {}; -- Mapped by (StandingID => StandingText)
local FactionStandingTextEnglish = { -- Lookup table for ByStanding text coloring
	"Hated", "Hostile", "Unfriendly", "Neutral", "Friendly", "Honored", "Revered", "Exalted"
};

local factionGroup = UnitFactionGroup("player");

local db;
local defaults = {
	profile = {
		Watch = {
			Bliz = false,
			Index = 1,
			FactionID = (factionGroup == "Horde" and 76 or 72),
		},
		DataStream = {
			Color = "Default",
			Show = {
				Name = true,
				Standing = true,
				Remaining = true,
			},
		},
		Colors = {
			Specific = {
				r = 1.0,
				g = 1.0,
				b = 1.0,
			},
			ByStanding = {
				Hated = {
					r = 0.80,
					g = 0.30,
					b = 0.22,
				},
				Hostile = {
					r = 0.80,
					g = 0.30,
					b = 0.22,
				},
				Unfriendly = {
					r = 0.75,
					g = 0.27,
					b = 0.00,
				},
				Neutral = {
					r = 0.9,
					g = 0.7,
					b = 0.0,
				},
				Friendly = {
					r = 0.0,
					g = 0.6,
					b = 0.1,
				},
				Honored = {
					r = 0.0,
					g = 0.6,
					b = 0.1,
				},
				Revered = {
					r = 0.0,
					g = 0.6,
					b = 0.1,
				},
				Exalted = {
					r = 0.0,
					g = 0.6,
					b = 0.1,
				},
			},
		},
		Tooltip = {
			Font = {
				Name = LSM:GetDefault("font"),
				Size = 12,
				Outline = " ",
			},
			Color = {
				Name = "Default",
				Progress = "Default",
				Standing = "Default",
				Remaining = "Default",
			},
			Backdrop = {
				Background = "Blizzard Tooltip",
				Border = "Blizzard Tooltip",
				BackgroundColor = {
					r = 0.09,
					g = 0.09,
					b = 0.19,
					a = 1.0,
				},
				BorderColor = {
					r = 1.0,
					g = 1.0,
					b = 1.0,
					a = 1.0,
				},
				Inset = 5,
				Tile = true,
				TileSize = 32,
				EdgeSize = 16,
			},
			Show = {
				Name = true,
				Progress = true,
				Standing = true,
				Remaining = true,
			},
			List = {
				Sort = "IndexDesc",
				SkipHeaders = true,
				Filter = {
					Hated = true,
					Hostile = true,
					Unfriendly = true,
					Neutral = true,
					Friendly = true,
					Honored = true,
					Revered = true,
					Exalted = true,
				},
			},
		},
	},
};

function addon:OnInitialize()
	self.Vars = LibStub("AceDB-3.0"):New("FactionBrokerDB", defaults);
	
	self.Vars.RegisterCallback(self, "OnProfileChanged", "UpdateDB");
	self.Vars.RegisterCallback(self, "OnProfileCopied", "UpdateDB");
	self.Vars.RegisterCallback(self, "OnProfileReset", "UpdateDB");
	
	db = self.Vars.profile;
	
	self:SetupBlizOptions();
	self:SetupFactionData();
	self:SetupFrame();
	
	LibStub("LibDataBroker-1.1"):NewDataObject("FactionBroker", self.LDBObject);
end

function addon:OnEnable()
	self:RegisterEvent("UPDATE_FACTION", "UpdateFaction");
	self:RegisterEvent("UNIT_FACTION", "UpdateLDBText");
end

function addon:OnDisable()
	self:UnregisterAllEvents();
end

function addon:SetupBlizOptions()
	--[===[ -- @todo Enable when the Options module gets added
	local frame, f;
	
	frame = CreateFrame("Frame", nil, UIParent);
	f = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate");
	
	frame.name = "FactionBroker";
	
	f:SetScript("OnClick", function(button)
		HideUIPanel(InterfaceOptionsFrame);
		HideUIPanel(GameMenuFrame);
		addon:OpenConfig();
	end);
	f:SetText("Open config window");
	f:SetPoint("CENTER");
	
	InterfaceOptions_AddCategory(frame);
	--]===]
	
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("FactionBroker", self.Options);
	
	self.OptionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("FactionBroker", "FactionBroker", nil);
	
	self.Options.args.Profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.Vars);
end

function addon:SetupFactionData()
	self.FactionList = {};
	
	local sex = UnitSex("player");
	local label = "FACTION_STANDING_LABEL%d";
	if sex == GENDER_FEMALE then
		label = label .. "_FEMALE";
	end
	
	for i = 1, 8 do
		FactionStandingText[i] = _G[string.format(label, i)];
	end
end

addon.Cache = {
	CheckForFactionData = function(cache, factionID)
		return cache.Factions[factionID] ~= nil;
	end,
	GetFactionData = function(cache, factionID)
		if not cache:CheckForFactionData(factionID) then
			return cache:CreateFactionData(factionID);
		end
		return cache.Factions[factionID];
	end,
	CreateFactionData = function(cache, factionID)
		if not cache:CheckForFactionData(factionID) then
			cache.Factions[factionID] = {};
		end
		return cache:GetFactionData(factionID);
	end,
	FindFactionData = function(cache, factionID)
		if not cache:CheckForFactionData(factionID) then
			addon:UpdateFactionList();
		end
		return cache:GetFactionData(factionID);
	end,
	Factions = {},
};

function addon:UpdateFactionList()
	wipe(self.FactionList);
	
	local factionData;
	for i = 1, GetNumFactions() do
		factionData = self:FetchFactionData(i);
		
		if factionData and db.Tooltip.List.Filter[ FactionStandingTextEnglish[factionData.StandingID] ] then
			if factionData.HasRep or (factionData.IsHeader and not db.Tooltip.List.SkipHeaders) then
				table.insert(self.FactionList, factionData);
			end
		end
	end
	
	table.sort(self.FactionList, self.ListSortFuncs[db.Tooltip.List.Sort]);
end

function addon:FetchFactionData(index)
	local factionData;
	local name, description, standingID, barMin, barMax, barValue, atWarWith, canToggleAtWar, isHeader, isCollapsed, hasRep, isWatched, isChild, factionID = GetFactionInfo(index);
	
	if not factionID then
		return;
	end
	
	-- Workaround to some factions having incorrect hasRep
	-- See FrameXML/ReputationFrame.lua:114
	hasRep = hasRep or not isHeader;
	
	local friendID, friendRep, friendMaxRep, friendText, friendTexture, friendTextLevel, friendThresh = GetFriendshipReputationByID(factionID);
	
	local standingText = nil;
	
	factionData = self.Cache:GetFactionData(factionID);
	
	if friendID then
		-- From FrameXML/ReputationFrame.xml:174
		standingText = friendTextLevel;
		--barMax = friendMaxRep - friendThresh; --
		barMax = 8400;
		barValue = barValue - friendThresh;
		barMin = 0;
		factionData.IsFriend = true;
		factionData.FriendID = friendID;
		factionData.Progress = friendRep;
	else
		factionData.IsFriend = false;
		standingText = FactionStandingText[standingID];
		factionData.Progress = barValue;
	end
	
	factionData.Name = name;
	factionData.StandingID = standingID;
	factionData.StandingText = standingText;
	factionData.Min = barMin;
	factionData.Max = barMax;
	factionData.Value = barValue;
	factionData.IsHeader = isHeader;
	factionData.HasRep = hasRep;
	factionData.FactionID = factionID;
	factionData.Index = index;
	factionData.BlizWatched = isWatched;
	
	return factionData;
end

function addon:FindWatchedFaction()
	self:UpdateFactionList();
	
	if db.Watch.Bliz then
		for k, v in pairs(self.Cache.Factions) do
			if v.BlizWatched then
				return v;
			end
		end
	else
		if self.Cache:CheckForFactionData(db.Watch.FactionID) then
			return self.Cache:GetFactionData(db.Watch.FactionID);
		else
			-- Faction info isn't ready yet
			return nil;
		end
	end
end

function addon:UpdateFactionInfo(index)
	local name, _, standing, min, max, value, _, _, header, _, hasRep, _, _ = GetFactionInfo(index);
	
	self.FactionInfo.Index = index;
	self.FactionInfo.Name = name;
	self.FactionInfo.Standing = standing;
	self.FactionInfo.Min = min;
	self.FactionInfo.Max = max;
	self.FactionInfo.Value = value;
	self.FactionInfo.IsHeader = header;
	self.FactionInfo.HasRep = hasRep;
	
	return self.FactionInfo;
end

function addon:UpdateDB(event, database)
	db = database.profile;
end

addon.Options = {
	type = 'group',
	name = "FactionBroker",
	get = function(info)
		return db[ info[#info] ];
	end,
	set = function(info, value)
		db[ info[#info] ] = value;
		addon:UpdateLDBText();
	end,
	childGroups = "tab",
	args = {
		WatchBliz = {
			type = 'toggle',
			order = 1,
			name = L["Use default UI watched faction"],
			desc = L["Watch the faction marked as |cffffd200Show as Experience Bar|r in the Blizzard Reputation window."],
			width = 'double',
			get = function(info)
				return db.Watch.Bliz;
			end,
			set = function(info, value)
				db.Watch.Bliz = value;
				addon:UpdateFaction();
			end
		},
		DataStream = {
			type = 'group',
			order = 2,
			name = L["DataBroker"],
			desc = L["DataBroker display options"],
			get = function(info)
				return db.DataStream[ info[#info] ];
			end,
			set = function(info, value)
				db.DataStream[ info[#info] ] = value;
				addon:UpdateLDBText();
			end,
			args = {
				Show = {
					type = 'group',
					order = 1,
					guiInline = true,
					name = L["Output control"],
					desc = L["Control what is shown in the DataBroker output."],
					get = function(info)
						return db.DataStream.Show[ info[#info] ];
					end,
					set = function(info, value)
						db.DataStream.Show[ info[#info] ] = value;
						addon:UpdateLDBText();
					end,
					args = {
						Name = {
							type = 'toggle',
							order = 1,
							name = L["Name"],
							desc = L["Show the faction's name in the DataBroker output."],
						},
						Standing = {
							type = 'toggle',
							order = 2,
							name = L["Standing"],
							desc = L["Show your standing with the faction in the DataBroker output."],
						},
						Remaining = {
							type = 'toggle',
							order = 3,
							name = L["Remaining"],
							desc = L["Show the amount remaining until the next standing level in the DataBroker output."],
						},
					},
				},
				Color = {
					type = 'select',
					order = 2,
					name = L["Text color scheme"],
					desc = L["Color scheme to use when preparing the DataBroker output. Each color scheme may be adjusted in the |cffffd200Color|r tab.\n\n|cffffd200Default|r: Do not specify a color. The DataBroker display addon will determine what color to use.\n\n|cffffd200Specific|r: User defined color used regardless of faction standing.\n\n|cffffd200By faction standing|r: Color the text based on your standing with the watched faction."],
					values = {
						Default = L["Default"],
						Specific = L["Specific"],
						ByStanding = L["By faction standing"],
					},
					
				},
			},
		},
		Colors = {
			type = 'group',
			order = 3,
			name = L["Color scheme"],
			desc = L["Color scheme options"],
			get = function(info)
				local t = db.Colors[ info[#info] ];
				return t.r, t.g, t.b, (t.a or 1.0);
			end,
			set = function(info, r, g, b, a)
				local t = db.Colors[ info[#info] ];
				t.r, t.g, t.b, t.a = r, g, b, a;
			end,
			args = {
				Specific = {
					type = 'color',
					order = 1,
					name = L["Specific color scheme"],
					desc = L["Color used with the |cffffd200Specific|r color scheme."],
					hasAlpha = false,
				},
				ByStanding = {
					type = 'group',
					order = 2,
					name = L["By standing color scheme"],
					guiInline = true,
					get = function(info)
						local t = db.Colors.ByStanding[ info[#info] ];
						return t.r, t.g, t.b, (t.a or 1.0);
					end,
					set = function(info, r, g, b, a)
						local t = db.Colors.ByStanding[ info[#info] ];
						t.r, t.g, t.b, t.a = r, g, b, a;
					end,
					args = {
						Exalted = {
							type = 'color',
							order = 1,
							name = FACTION_STANDING_LABEL8,
							hasAlpha = false,
						},
						Revered = {
							type = 'color',
							order = 2,
							name = FACTION_STANDING_LABEL7,
							hasAlpha = false,
						},
						Honored = {
							type = 'color',
							order = 3,
							name = FACTION_STANDING_LABEL6,
							hasAlpha = false,
						},
						Friendly = {
							type = 'color',
							order = 4,
							name = FACTION_STANDING_LABEL5,
							hasAlpha = false,
						},
						Neutral = {
							type = 'color',
							order = 5,
							name = FACTION_STANDING_LABEL4,
							hasAlpha = false,
						},
						Unfriendly = {
							type = 'color',
							order = 6,
							name = FACTION_STANDING_LABEL3,
							hasAlpha = false,
						},
						Hostile = {
							type = 'color',
							order = 7,
							name = FACTION_STANDING_LABEL2,
							hasAlpha = false,
						},
						Hated = {
							type = 'color',
							order = 8,
							name = FACTION_STANDING_LABEL1,
							hasAlpha = false,
						},
					},
				},
			},
		},
		Tooltip = {
			type = 'group',
			order = 4,
			name = L["List"],
			desc = L["Faction list frame options"],
			get = function(info)
				return db.Tooltip[ info[#info] ];
			end,
			set = function(info, value)
				db.Tooltip[ info[#info] ] = value;
			end,
			childGroups = "tab",
			args = {
				Font = {
					type = 'group',
					order = 1,
					name = L["Text"],
					desc = L["Text font options"],
					get = function(info)
						return db.Tooltip.Font[ info[#info] ];
					end,
					set = function(info, value)
						db.Tooltip.Font[ info[#info] ] = value;
					end,
					args = {
						Name = {
							type = 'select',
							order = 1,
							name = L["Font"],
							dialogControl = "LSM30_Font",
							values = AceGUIWidgetLSMlists.font,
						},
						Size = {
							type = 'range',
							order = 2,
							name = L["Text size"],
							min = 6,
							max = 32,
							step = 1,
						},
						Outline = {
							type = "select",
							order = 3,
							name = L["Text outline"],
							values = {
								[" "] = L["None"],
								["OUTLINE"] = L["Outline"],
								["THICKOUTLINE"] = L["Thick outline"],
							},
						},
					},
				},
				Color = {
					type = 'group',
					order = 2,
					name = L["Color"],
					desc = L["Color scheme options"],
					get = function(info)
						return db.Tooltip.Color[ info[#info] ];
					end,
					set = function(info, value)
						db.Tooltip.Color[ info[#info] ] = value;
					end,
					args = {
						Name = {
							type = 'select',
							order = 1,
							name = L["Name"],
							desc = L["Color scheme used to color text in the |cffffd200Name|r column."],
							values = {
								Default = L["Default"],
								Specific = L["Specific"],
								ByStanding = L["By faction standing"],
							},
						},
						Progress = {
							type = 'select',
							order = 2,
							name = L["Progress"],
							desc = L["Color scheme used to color text in the |cffffd200Progress|r column."],
							values = {
								Default = L["Default"],
								Specific = L["Specific"],
								ByStanding = L["By faction standing"],
							},
						},
						Standing = {
							type = 'select',
							order = 3,
							name = L["Standing"],
							desc = L["Color scheme used to color text in the |cffffd200Standing|r column."],
							values = {
								Default = L["Default"],
								Specific = L["Specific"],
								ByStanding = L["By faction standing"],
							},
						},
						Remaining = {
							type = 'select',
							order = 4,
							name = L["Remaining"],
							desc = L["Color scheme used to color text in the |cffffd200Remaining|r column."],
							values = {
								Default = L["Default"],
								Specific = L["Specific"],
								ByStanding = L["By faction standing"],
							},
						},
					},
				},
				Backdrop = {
					type = 'group',
					order = 3,
					name = L["Backdrop"],
					desc = L["Background and border appearance options"],
					get = function(info)
						return db.Tooltip.Backdrop[ info[#info] ];
					end,
					set = function(info, value)
						db.Tooltip.Backdrop[ info[#info] ] = value;
						addon:UpdateFrameArt();
					end,
					args = {
						Background = {
							type = 'select',
							order = 1,
							name = L["Background texture"],
							desc = L["Image used as the background for the frame."],
							dialogControl = "LSM30_Background",
							values = AceGUIWidgetLSMlists.background,
							width = 'double',
						},
						BackgroundColor = {
							type = 'color',
							order = 2,
							name = L["Background color"],
							desc = L["Color used to tint the background image.\n\n|cffffd200Note|r: Dark images will not change drastically."],
							hasAlpha = true,
							get = function(info)
								local t = db.Tooltip.Backdrop.BackgroundColor;
								return t.r, t.g, t.b, t.a;
							end,
							set = function(info, r, g, b, a)
								local t = db.Tooltip.Backdrop.BackgroundColor;
								t.r, t.g, t.b, t.a = r, g, b, a;
								addon:UpdateFrameArt();
							end,
						},
						Inset = {
							type = 'range',
							order = 3,
							name = L["Background inset"],
							desc = L["Distance between the border and the background image."],
							min = 0,
							max = 32,
							step = 1,
						},
						Tile = {
							type = 'toggle',
							order = 4,
							name = L["Tile background"],
							desc = L["Tile the background image. When this options is NOT selected, the background image will be stretched to fit the faction list frame."],
						},
						TileSize = {
							type = 'range',
							order = 5,
							name = L["Tile size"],
							desc = L["Size of the square tile used to tile background image."],
							disabled = function()
								return not db.Tooltip.Backdrop.Tile;
							end,
							min = 2,
							max = 32,
							step = 1,
						},
						Border = {
							type = 'select',
							order = 6,
							name = L["Border texture"],
							desc = L["Image used as the border for the frame."],
							dialogControl = "LSM30_Border",
							values = AceGUIWidgetLSMlists.border,
							width = 'double',
						},
						BorderColor = {
							type = 'color',
							order = 7,
							name = L["Border color"],
							desc = L["Color used to tint the border image.\n\n|cffffd200Note|r: Dark images will not change drastically."],
							hasAlpha = true,
							get = function(info)
								local t = db.Tooltip.Backdrop.BorderColor;
								return t.r, t.g, t.b, t.a;
							end,
							set = function(info, r, g, b, a)
								local t = db.Tooltip.Backdrop.BorderColor;
								t.r, t.g, t.b, t.a = r, g, b, a;
								addon:UpdateFrameArt();
							end,
						},
						EdgeSize = {
							type = 'range',
							order = 8,
							name = L["Border thickness"],
							min = 2,
							max = 32,
							step = 1,
						},
					},
				},
				Show = {
					type = 'group',
					order = 4,
					name = L["Columns"],
					desc = L["Options to show/hide columns"],
					get = function(info)
						return db.Tooltip.Show[ info[#info] ];
					end,
					set = function(info, value)
						db.Tooltip.Show[ info[#info] ] = value;
					end,
					args = {
						Name = {
							type = 'toggle',
							order = 1,
							name = L["Faction name"],
							desc = L["Show the |cffffd200Name|r column"],
							width = 'double',
						},
						Progress = {
							type = 'toggle',
							order = 1,
							name = L["Progress"],
							desc = L["Show the |cffffd200Progress|r column"],
							width = 'double',
						},
						Standing = {
							type = 'toggle',
							order = 1,
							name = L["Standing"],
							desc = L["Show the |cffffd200Standing|r column"],
							width = 'double',
						},
						Remaining = {
							type = 'toggle',
							order = 1,
							name = L["Remaining"],
							desc = L["Show the |cffffd200Remaining|r column"],
							width = 'double',
						},
						
					},
				},
				List = {
					type = 'group',
					order = 5,
					name = L["List"],
					desc = L["Sort and exclude factions in the list."],
					get = function(info)
						return db.Tooltip.List[ info[#info] ];
					end,
					set = function(info, value)
						db.Tooltip.List[ info[#info] ] = value;
					end,
					args = {
						SkipHeaders = {
							type = 'toggle',
							order = 1,
							name = L["Skip headers"],
							desc = L["Skip headers with reputation bars (Alliance Vanguard, Horder Expedition, etc)"],
							width = 'double',
						},
						Sort = {
							type = 'select',
							order = 2,
							name = L["Sort method"],
							desc = L["Metric used to sort the faciton list."],
							values = {
								IndexAsc = L["Faction index, Ascending"],
								IndexDesc = L["Faction index, Descending"],
								AlphaAsc = L["Alphabetically, Ascending"],
								AlphaDesc = L["Alphabetically, Descending"],
								ProgressAsc = L["Total progress, Ascending"],
								ProgressDesc = L["Total progress, Descending"],
								StandingAsc = L["Standing level, Ascending"],
								StandingDesc = L["Standing level, Descending"],
								RemainingAsc = L["Remaining, Ascending"],
								RemaingingDesc = L["Remaining, Descending"],
							},
						},
						Filter = {
							type = 'group',
							order = 3,
							name = L["Filter by standing"],
							guiInline = true,
							get = function(info)
								return db.Tooltip.List.Filter[ info[#info] ];
							end,
							set = function(info, value)
								db.Tooltip.List.Filter[ info[#info] ] = value;
							end,
							args = {
								Exalted = {
									type = 'toggle',
									order = 1,
									name = FACTION_STANDING_LABEL8,
									desc = L["Show factions at this standing level"],
								},
								Revered = {
									type = 'toggle',
									order = 2,
									name = FACTION_STANDING_LABEL7,
									desc = L["Show factions at this standing level"],
								},
								Honored = {
									type = 'toggle',
									order = 3,
									name = FACTION_STANDING_LABEL6,
									desc = L["Show factions at this standing level"],
								},
								Friendly = {
									type = 'toggle',
									order = 4,
									name = FACTION_STANDING_LABEL5,
									desc = L["Show factions at this standing level"],
								},
								Neutral = {
									type = 'toggle',
									order = 5,
									name = FACTION_STANDING_LABEL4,
									desc = L["Show factions at this standing level"],
								},
								Unfriendly = {
									type = 'toggle',
									order = 6,
									name = FACTION_STANDING_LABEL3,
									desc = L["Show factions at this standing level"],
								},
								Hostile = {
									type = 'toggle',
									order = 7,
									name = FACTION_STANDING_LABEL2,
									desc = L["Show factions at this standing level"],
								},
								Hated = {
									type = 'toggle',
									order = 8,
									name = FACTION_STANDING_LABEL1,
									desc = L["Show factions at this standing level"],
								},
							},
						},
					},
				},
			},
		},
	},
};

function addon:OpenConfig()
	InterfaceOptionsFrame_OpenToCategory(self.OptionsFrame);
end

addon.LDBObject = {
	type = 'data source',
	icon = [[Interface\PVPFrame\PVP-Currency-]] .. factionGroup .. [[.blp]],
	label = "FactionBroker",
	text = "Loading...",
	tocname = "FactionBroker",
	OnClick = function(self, button)
		if button == "LeftButton" then
			ToggleCharacter("ReputationFrame");
		end
		if button == "RightButton" then
			addon:OpenConfig();
		end
	end,
	OnEnter = function(button)
		addon:ShowFrame(button);
	end,
	OnLeave = function(button)
		addon:HideFrame();
	end,
};

function addon:UpdateFaction()
	local factionData;
	if db.Watch.Bliz then
		factionData = self:FindWatchedFaction();
		if factionData then
			self:SetWatchedFaction(factionData);
		end
	end
	
	self:UpdateLDBText();
end

function addon:UpdateLDBText()
	local factionData = self:FindWatchedFaction();
	local str = "{2}/{3}";
	
	if db.DataStream.Show.Remaining then
		str = str .. "({5})";
	end
	
	if db.DataStream.Show.Name then
		str = "{1}: " .. str;
	end
	
	if db.DataStream.Show.Standing then
		str = str .. " ({4})";
	end
	
	local text = L["Loading..."];
	
	if factionData then
		local breadth = factionData.Max - factionData.Min;
		local value = factionData.Value - factionData.Min;
		local standingText = factionData.StandingText;
		local remain = factionData.Max - factionData.Value;
		
		text = self:Format(str, factionData.Name, value, breadth, standingText, remain);
		
		local color = self:LookupColor("LDB", factionData);
		
		if color then
			text = string.format("|cff%02x%02x%02x%s|r", color.r * 255, color.g * 255, color.b * 255, text);
		end
	end
	
	self.LDBObject.text = text;
end

function addon:SetupFrame()
	local frame = _G.FactionBrokerFrame;
	
	for k, v in pairs(self.FrameFuncs) do
		frame[k] = v;
	end
	
	frame.Lines = {};
	frame.TextFrames = {};
	frame.Backdrop = {
		insets = {
			top = 0,
			bottom = 0,
			left = 0,
			right = 0,
		},
	};
	
	frame:SetScript("OnShow", frame.OnShow);
	frame:SetScript("OnHide", frame.OnHide);
	frame:SetScript("OnLeave", frame.OnLeave);
	
	frame.Footer:SetText(L["|cffffff00Left click|r: Open the reputation frame\n|cffffff00Right click|r: Open FactionBroker config window"]);
	
	self.Frame = frame;
end

function addon:LookupColor(fieldName, factionData)
	local scheme;
	if fieldName == "LDB" then
		scheme = db.DataStream.Color;
	elseif string.match(fieldName, '^Tooltip') then
		local section = string.match(fieldName, '^Tooltip%.(.*)$');
		scheme = db.Tooltip.Color[section];
	else
		print(string.format("FactionBroker:LookupColor: unknown fieldName (%s).", tostring(fieldName)));
		return;
	end
	
	if self.ColorModeFuncs[scheme] then
		return self.ColorModeFuncs[scheme](factionData);
	else
		print(string.format("FactionBroker:LookupColor: unknown color scheme (%s).", tostring(scheme)));
		return;
	end
end

addon.ColorModeFuncs = {
	Specific = function()
		return db.Colors.Specific;
	end,
	Default = function()
		-- Let the LDB display choose a color
		return nil;
	end,
	ByStanding = function(factionData)
		return db.Colors.ByStanding[ FactionStandingTextEnglish[factionData.StandingID] ];
	end
};

local NumTextFrames = 0;

addon.FrameFuncs = {
	OnShow = function(frame)
		frame:UpdateArt();
		frame:SetScript("OnUpdate", frame.OnUpdate);
	end,
	OnHide = function(frame)
		frame:SetScript("OnUpdate", nil);
		frame.InChild = false;
	end,
	OnLeave = function(frame)
		if not frame:IsMouseOver(1, -1, -1, 1) then
			frame:Hide();
		else
			frame.InChild = true;
		end
	end,
	OnUpdate = function(frame)
		-- Here because "OnLeave" wont fire if you move the mouse quickly off of a child with mouse motion enabled.
		if frame.InChild and not frame:IsMouseOver(1, -1, -1, 1) then
			frame:Hide();
		end
	
		frame:ClearLines();
		local head = frame.Container.Header;
		local Name_MaxWidth, Progress_MaxWidth, Standing_MaxWidth, Remaining_MaxWidth = head.Name:GetStringWidth(), head.Progress:GetStringWidth(), head.Standing:GetStringWidth(), head.Remaining:GetStringWidth();
		local height = (2 * db.Tooltip.Backdrop.Inset) + frame.Title:GetHeight() + head:GetHeight() + frame.Footer:GetHeight() + 15;
		local item_MaxHeight = 0;
		
		head.Name:SetWidth(100);
		head.Progress:SetWidth(100);
		head.Standing:SetWidth(100);
		head.Remaining:SetWidth(100);
		
		head.Name:Show();
		head.Progress:Show();
		head.Standing:Show();
		head.Remaining:Show();
		
		local last = head;
		local container = frame.Container;
		local color;
		local standingText;
		local factionData;
		
		-- Sorting is done in :UpdateFactionList
		for k, v in pairs(addon.FactionList) do
			frame:AddLine(v);
		end
		
		for k, v in pairs(frame.Lines) do
			item_MaxHeight = 0;
			
			factionData = v.FactionData;
			
			standingText = factionData.StandingText;
			
			if db.Tooltip.Show.Name then
				if v.Name:GetStringWidth() > Name_MaxWidth then
					Name_MaxWidth = v.Name:GetStringWidth();
				end
				if v.Name:GetStringHeight() > item_MaxHeight then
					item_MaxHeight = v.Name:GetStringHeight();
				end
				v.Name:SetPoint("LEFT", head.Name);
				v.Name:SetPoint("RIGHT", head.Name);
				color = addon:LookupColor("Tooltip.Name", factionData);
				if color then
					v.Name:SetTextColor(color.r, color.g, color.b);
				end
			else
				v.Name:Hide();
			end
			
			if db.Tooltip.Show.Progress then
				if v.Progress:GetStringWidth() > Progress_MaxWidth then
					Progress_MaxWidth = v.Progress:GetStringWidth();
				end
				if v.Progress:GetStringHeight() > item_MaxHeight then
					item_MaxHeight = v.Progress:GetStringHeight();
				end
				v.Progress:SetPoint("LEFT", head.Progress);
				v.Progress:SetPoint("RIGHT", head.Progress);
				color = addon:LookupColor("Tooltip.Progress", factionData);
				if color then
					v.Progress:SetTextColor(color.r, color.g, color.b);
				end
			else
				v.Progress:Hide();
			end
			
			if db.Tooltip.Show.Standing then
				if v.Standing:GetStringWidth() > Standing_MaxWidth then
					Standing_MaxWidth = v.Standing:GetStringWidth();
				end
				if v.Standing:GetStringHeight() > item_MaxHeight then
					item_MaxHeight = v.Standing:GetStringHeight();
				end
				v.Standing:SetPoint("LEFT", head.Standing);
				v.Standing:SetPoint("RIGHT", head.Standing);
				color = addon:LookupColor("Tooltip.Standing", factionData);
				if color then
					v.Standing:SetTextColor(color.r, color.g, color.b);
				end
			else
				v.Standing:Hide();
			end
			
			if db.Tooltip.Show.Remaining then
				if v.Remaining:GetStringWidth() > Remaining_MaxWidth then
					Remaining_MaxWidth = v.Remaining:GetStringWidth();
				end
				if v.Remaining:GetStringHeight() > item_MaxHeight then
					item_MaxHeight = v.Remaining:GetStringHeight();
				end
				v.Remaining:SetPoint("LEFT", head.Remaining);
				v.Remaining:SetPoint("RIGHT", head.Remaining);
				color = addon:LookupColor("Tooltip.Remaining", factionData);
				if color then
					v.Remaining:SetTextColor(color.r, color.g, color.b);
				end
			else
				v.Remaining:Hide();
			end
			
			v:Show();
		
			v:SetPoint("TOP", last, "BOTTOM", 0, -5);
			v:SetPoint("LEFT", container, "LEFT", 5, 0);
			v:SetPoint("RIGHT", container, "RIGHT", -5, 0);
			
			height = height + item_MaxHeight + 5;
			
			v:SetHeight(item_MaxHeight);
			
			last = v;
		end
		
		last = nil;
		
		local width = 30;
		
		if db.Tooltip.Show.Name then
			head.Name:SetPoint("LEFT", container, "LEFT", 5, 0);
			head.Name:SetWidth(Name_MaxWidth);
			width = width + Name_MaxWidth + 5;
			last = head.Name;
		else
			head.Name:Hide();
		end
		
		if db.Tooltip.Show.Progress then
			if not last then
				head.Progress:SetPoint("LEFT", container, "LEFT", 5, 0);
			else
				head.Progress:SetPoint("LEFT", last, "RIGHT", 5, 0);
			end
			head.Progress:SetWidth(Progress_MaxWidth);
			width = width + Progress_MaxWidth + 5;
			last = head.Progress;
		else
			head.Progress:Hide();
		end
		
		if db.Tooltip.Show.Standing then
			if not last then
				head.Standing:SetPoint("LEFT", container, "LEFT", 5, 0);
			else
				head.Standing:SetPoint("LEFT", last, "RIGHT", 5, 0);
			end
			head.Standing:SetWidth(Standing_MaxWidth);
			width = width + Standing_MaxWidth + 5;
			last = head.Standing;
		else
			head.Standing:Hide();
		end
		
		if db.Tooltip.Show.Remaining then
			if not last then
				head.Remaining:SetPoint("LEFT", container, "LEFT", 5, 0);
			else
				head.Remaining:SetPoint("LEFT", last, "RIGHT", 5, 0);
			end
			head.Remaining:SetWidth(Remaining_MaxWidth);
			width = width + Remaining_MaxWidth + 5;
		else
			head.Remaining:Hide();
		end
		
		frame:SetWidth(width);
		frame:SetHeight(height);
	end,
	GetTextFrame = function(frame)
		local f = next(frame.TextFrames);
		-- Make a new frame
		if not f then
			NumTextFrames = NumTextFrames + 1;
			f = CreateFrame("Button", "FactionBrokerTextFrame" .. NumTextFrames, frame.Container, "FactionBrokerLineTemplate");
		else
			frame.TextFrames[f] = nil;
		end
		
		local font, size, flags = LSM:Fetch('font', db.Tooltip.Font.Name), db.Tooltip.Font.Size, db.Tooltip.Font.Outline;
		
		f:SetPoint("LEFT", frame.Container, "LEFT", 5, 0);
		f:SetPoint("RIGHT", frame.Container, "RIGHT", -5, 0);
		
		f:SetHeight(25);
		f:SetWidth(500);
		
		f.Name:ClearAllPoints();
		f.Name:SetPoint("TOP", f, 0, 0);
		f.Name:SetPoint("LEFT", f, 5, 0);
		f.Name:SetHeight(size);
		f.Name:SetWidth(200);
		f.Name:Show();
		f.Name:SetFont(font, size, flags);
		f.Name:SetTextColor(1.0, 1.0, 1.0, 1.0);
		
		f.Progress:ClearAllPoints();
		f.Progress:SetPoint("TOP", f, 0, 0);
		f.Progress:SetPoint("LEFT", f.Name, "RIGHT", 5, 0);
		f.Progress:SetHeight(size);
		f.Progress:SetWidth(100);
		f.Progress:Show();
		f.Progress:SetFont(font, size, flags);
		f.Progress:SetTextColor(1.0, 1.0, 1.0, 1.0);
		
		f.Standing:ClearAllPoints();
		f.Standing:SetPoint("TOP", f, 0, 0);
		f.Standing:SetPoint("LEFT", f.Progress, "RIGHT", 5, 0);
		f.Standing:SetHeight(size);
		f.Standing:SetWidth(100);
		f.Standing:Show();
		f.Standing:SetFont(font, size, flags);
		f.Standing:SetTextColor(1.0, 1.0, 1.0, 1.0);
		
		f.Remaining:ClearAllPoints();
		f.Remaining:SetPoint("TOP", f, 0, 0);
		f.Remaining:SetPoint("LEFT", f.Standing, "RIGHT", 5, 0);
		f.Remaining:SetHeight(size);
		f.Remaining:SetWidth(100);
		f.Remaining:Show();
		f.Remaining:SetFont(font, size, flags);
		f.Remaining:SetTextColor(1.0, 1.0, 1.0, 1.0);
		
		return f;
	end,
	AddLine = function(frame, factionData)
		local f = frame:GetTextFrame();
		f.FactionData = factionData;
				
		if not db.Tooltip.Show.Name then
			f.Name:Hide();
			f.Name:SetText("");
			f.Name:SetWidth(0);
		else
			f.Name:SetText(factionData.Name);
		end
		
		if not db.Tooltip.Show.Progress then
			f.Progress:Hide();
			f.Progress:SetText("");
			f.Progress:SetWidth(0);
		else
			f.Progress:SetFormattedText("%d/%d", factionData.Value - factionData.Min, factionData.Max - factionData.Min);
		end
		
		if not db.Tooltip.Show.Remaining then
			f.Remaining:Hide();
			f.Remaining:SetText("");
			f.Remaining:SetWidth(0);
		else
			f.Remaining:SetText(tostring(factionData.Max - factionData.Value));
		end
		
		if not db.Tooltip.Show.Standing then
			f.Standing:Hide();
			f.Standing:SetText("");
			f.Standing:SetWidth(0);
		else
			f.Standing:SetText(factionData.StandingText);
		end
		
		table.insert(frame.Lines, f);
	end,
	ClearLines = function(frame)
		for k, v in pairs(frame.Lines) do
			frame.Lines[k] = nil;
			v:Hide();
			v:ClearAllPoints();
			frame.TextFrames[v] = true;
		end
	end,
	UpdateArt = function(frame)
		-- Update frame backdrop and border
		do
			local backdrop = frame.Backdrop;
			local insets = backdrop.insets;
			
			backdrop.bgFile = LSM:Fetch("background", db.Tooltip.Backdrop.Background);
			backdrop.edgeFile = LSM:Fetch("border", db.Tooltip.Backdrop.Border);
			backdrop.tile = db.Tooltip.Backdrop.Tile;
			backdrop.tileSize = db.Tooltip.Backdrop.TileSize;
			backdrop.edgeSize = db.Tooltip.Backdrop.EdgeSize;
			
			for k, v in pairs(insets) do
				insets[k] = db.Tooltip.Backdrop.Inset;
			end
			
			frame:SetBackdrop(backdrop);
			
			local color = db.Tooltip.Backdrop.BackgroundColor;
			
			frame:SetBackdropColor(color.r, color.g, color.b, color.a);
			
			color = db.Tooltip.Backdrop.BorderColor;
			
			frame:SetBackdropBorderColor(color.r, color.g, color.b, color.a);
		end
		
		-- Update static text
		do
			local font, size, outline = LSM:Fetch("font", db.Tooltip.Font.Name), db.Tooltip.Font.Size, db.Tooltip.Font.Outline;
		
			frame.Title:SetHeight(50);
			frame.Title:SetFont(font, size, outline);
			frame.Title:SetHeight(frame.Title:GetStringHeight());
			frame.Title:SetPoint("TOP", frame, "TOP", 0, -db.Tooltip.Backdrop.Inset);
			
			frame.Footer:SetHeight(75);
			frame.Footer:SetFont(font, size, outline);
			frame.Footer:SetHeight(frame.Footer:GetStringHeight());
			frame.Footer:SetPoint("BOTTOM", frame, "BOTTOM", 0, 1.5 * db.Tooltip.Backdrop.Inset);
			
			local head = frame.Container.Header;
			
			head.Name:SetFont(font, size, outline);
			head.Name:SetHeight(size);
			
			head.Progress:SetFont(font, size, outline);
			head.Progress:SetHeight(size);
			
			head.Standing:SetFont(font, size, outline);
			head.Standing:SetHeight(size);
			
			head.Remaining:SetFont(font, size, outline);
			head.Remaining:SetHeight(size);
		end
		
		-- @TODO: Update other art assets
	end,
};

function addon:ShowFrame(anchor)
	local frame = self.Frame;
	
	do
		-- "Smart" anchor borrowed from LibQTip-1.0
		frame:ClearAllPoints();
		local x, y = anchor:GetCenter();
		-- The anchor frame isn't shown
		if not x or not y then
			-- So anchor to the parent frame
			frame:SetPoint("TOPLEFT", "BOTTOMLEFT");
		else
			local width, height = UIParent:GetWidth(), UIParent:GetHeight();
			local horiz, vert = "", "BOTTOM";
			if x > width * 2 / 3 then
				horiz = "RIGHT";
			elseif x < width / 3 then
				horiz = "LEFT";
			end
			if y > height / 2 then
				vert = "TOP";
			end
			frame:SetPoint(vert .. horiz, anchor, (vert == "TOP" and "BOTTOM" or "TOP") .. horiz);
		end
	end
	
	self:UpdateFactionList();
	
	frame:Show();
end

function addon:HideFrame()
	local frame = self.Frame;
	if not frame:IsMouseOver(1, -1, -1, 1) then
		frame:Hide();
	end
end

function addon:UpdateFrameArt()
	addon.Frame:UpdateArt();
end

addon.ListSortFuncs = {
	IndexAsc = function(a, b)
		return b.Index > a.Index;
	end,
	IndexDesc = function(a, b)
		return a.Index > b.Index;
	end,
	AlphaAsc = function(a, b)
		return a.Name < b.Name;
	end,
	AlphaDesc = function(a, b)
		return a.Name > b.Name;
	end,
	ProgressAsc = function(a, b)
		return a.Progress < b.Progress;
	end,
	ProgressDesc = function(a, b)
		return a.Progress > b.Progress;
	end,
	StandingAsc = function(a, b)
		return a.StandingID < b.StandingID;
	end,
	StandingDesc = function(a, b)
		return a.StandingID > b.StandingID;
	end,
	RemainingAsc = function(a, b)
		local left1, left2 = a.Max - a.Value, b.Max - b.Min;
		if left1 == left2 then
			return a.Max < b.Max;
		else
			return left1 < left2;
		end
	end,
	RemainingDesc = function(a, b)
		local left1, left2 = a.Max - a.Value, b.Max - b.Min;
		if left1 == left2 then
			return a.Max > b.Max;
		else
			return left1 > left2;
		end
	end,
};

function addon:SetWatchedFaction(factionData)
	if not factionData then
		return;
	end
	if db.Watch.Bliz then
		SetWatchedFactionIndex(factionData.Index);
	end
	
	db.Watch.FactionID = factionData.FactionID;
	
	self:UpdateLDBText();
end

do
	local format_args = {};

	function addon:Format(msg, ...)
		local limit = select("#", ...);
		for i = 1, limit do
			format_args[i] = select(i, ...);
		end
		
		local str = string.gsub(msg, "{(%d+)}", function(d)
			return tostring(format_args[tonumber(d)]);
		end);
		wipe(format_args);
		return str;
	end
end